# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import re
from typing import Optional, Union

from databind.core import Converter
from nr.utils.git import Git  # type: ignore
from packaging.version import Version as _Version

logger = logging.getLogger(__name__)


class Version(_Version):
  """ An extension of #packageing.version.Version which supports a
  commit-distance and commit SHA suffix in the format of `-X-gY` (where
  X is the distance and Y is the lowercase 7-character SHA sum). """

  commit_distance: Optional[int]

  def __init__(self, s: Union['Version', str]):
    if isinstance(s, Version):
      s = str(s)
    elif not isinstance(s, str):
      raise TypeError('expected Version or str, got {}'.format(type(s).__name__))
    match = re.match(r'(.*)-(\d+)-g([0-9a-f]{7})', s)
    if match:
      s = match.group(1)
      commit_distance = int(match.group(2))
      sha = match.group(3)
    else:
      commit_distance = None
      sha = None
    super().__init__(s)
    self.commit_distance = commit_distance
    self.sha = sha

  def __str__(self):
    s = super().__str__()
    if self.commit_distance and self.sha:
      s += '-{}-g{}'.format(self.commit_distance, self.sha)
    return s

  def __lt__(self, other):
    if super().__lt__(other):
      return True
    if super().__eq__(other):
      return (self.commit_distance or 0) < (other.commit_distance or 0)
    return False

  def __gt__(self, other):
    return other < self and other != self

  def __eq__(self, other):
    if super().__eq__(other) is True:
      return (self.commit_distance, self.sha) == (other.commit_distance, other.sha)
    return False

  def __ne__(self, other):
    return not (self == other)

  @property
  def pep440_compliant(self):
    return self.sha is None


def parse_version(version_string: str) -> Version:
  return Version(version_string)


def bump_version(version: Version, kind: str) -> Version:
  major, minor, patch, post = version.major, version.minor, version.micro, version.post
  if kind == 'post':
    if post is None:
      post = 1
    else:
      post += 1
  elif kind == 'patch':
    post = None
    patch += 1
  elif kind == 'minor':
    post = None
    patch = 0
    minor += 1
  elif kind == 'major':
    post = None
    patch = minor = 0
    major += 1
  else:
    raise ValueError('invalid kind: {!r}'.format(kind))
  string = '%s.%s.%s' % (major, minor, patch)
  if post:
    string += '.post' + str(post)
  return Version(string)


def get_commit_distance_version(repo_dir: str, version: Version, latest_tag: str) -> Optional[Version]:
  """
  This function creates a string which describes the version of the
  monorepo or package that includes the commit distance and SHA revision
  number.

  For a mono repository, the full commit distance is used. The same is true
  for a single package. For a package inside a mono repository that does not
  apply mono versioning, the packages' local commit distance is used.

  This is close to what `git describe --tags` does. An example version number
  generated by this function is: `0.1.0+24.gd9ade3f`. If the working state is
  dirty, `.dirty` will be appended to the local version.

  Notes:

  - If there is no commit distance from the *latest_tag* to the current
    state of the repository, this function returns None.
  - The version returned by this function is a PEP440 local version that
    cannot be used for packages when submitting them to PyPI.
  - If the tag for the version of *subject* does not exist on the repository,
    it will fall back to 0.0.0 as the version number which is treated as
    "the beginning of the repository", even if no tag for this version exists.

    Todo: We could try to find the previous tag for this subject and use that.
  """

  git = Git(repo_dir)
  dirty = git.has_diff()

  if git.rev_parse(latest_tag):
    distance = len(git.rev_list(latest_tag + '..HEAD'))
  else:
    logger.warning('tag "%s" does not exist', latest_tag)
    version = Version('0.0.0')
    distance = len(git.rev_list('HEAD'))

  if distance == 0:
    if dirty:
      return parse_version(str(version) + '+dirty')
    return None

  rev = git.rev_parse('HEAD')
  local = '+{}.g{}{}'.format(distance, rev[:7], '.dirty' if dirty else '')
  return parse_version(str(version) + local)


from . import registry


class VersionConverter(Converter):

  def from_python(self, value, context):
    return str(value)

  def to_python(self, value, context):
    return parse_version(value)


registry.register_converter(Version, VersionConverter())



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import logging
from typing import Iterable

import nr.fs  # type: ignore
from nr.stream import Stream  # type: ignore
from termcolor import colored

from shut.changelog.manager import ChangelogManager
from shut.commands.commons.bump import make_bump_command, VersionBumpData, VersionRef
from shut.commands.pkg.bump import PackageBumpData
from shut.model import MonorepoModel
from shut.model.version import get_commit_distance_version, Version
from shut.utils.io.virtual import VirtualFiles
from shut.utils.text import substitute_ranges
from . import mono
from .update import update_monorepo

logger = logging.getLogger(__name__)


class MonorepoBumpdata(VersionBumpData[MonorepoModel]):

  def update(self, new_version: Version) -> VirtualFiles:
    # We have to re-load the monorepo and package definitions from the files since
    # they have been updated by #bump_to_version(). This is a workaround to updating
    # the version selectors of inter-dependencies in memory.
    self.project.reload()

    vfiles = update_monorepo(self.obj, dry=self.args.dry, indent=1)
    if self.obj.release.single_version:
      for package in self.project.packages:
        vfiles.update(
          PackageBumpData(self.args, self.project, package).update(new_version),
          os.path.relpath(package.get_directory(), self.obj.get_directory()))

    return vfiles

  def get_version_refs(self) -> Iterable[VersionRef]:
    yield from super().get_version_refs()
    if self.obj.release.single_version:
      for package in self.project.packages:
        yield from PackageBumpData(self.args, self.project, package).get_version_refs()

  def bump_to_version(self, target_version: Version) -> Iterable[str]:
    changed_files = list(super().bump_to_version(target_version))

    if not self.obj.release.single_version:
      return changed_files

    inter_deps = list(self.obj.get_inter_dependencies())
    if not inter_deps:
      return changed_files

    print()
    print(f'bumping {len(inter_deps)} mono repository inter-dependency(-ies)')
    for filename, refs in Stream.groupby(inter_deps, lambda d: d.filename, collect=list):
      print(f'  {colored(nr.fs.rel(filename), "cyan")}:')

      with open(filename) as fp:
        content = fp.read()

      for ref in refs:
        value = content[ref.version_start:ref.version_end]
        print(f'    {ref.package_name} {value} → ^{target_version}')

      content = substitute_ranges(
        content,
        ((ref.version_start, ref.version_end, f'^{target_version}') for ref in refs),
      )
      if not self.args.dry:
        with open(filename, 'w') as fp:
          fp.write(content)

      changed_files.append(filename)

    if self.args.tag and inter_deps and self.args.skip_update:
      logger.warning('bump requires an update in order to automatically tag')

    return changed_files

  def get_snapshot_version(self) -> Version:
    return get_commit_distance_version(
      self.obj.get_directory(),
      self.obj.version,
      self.obj.get_tag(self.obj.version)) or self.obj.version

  def get_changelog_managers(self) -> Iterable[ChangelogManager]:
    yield from super().get_changelog_managers()
    if self.obj.release.single_version:
      for package in self.project.packages:
        yield from PackageBumpData(self.args, self.project, package).get_changelog_managers()


mono.command()(make_bump_command(MonorepoBumpdata, MonorepoModel))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import contextlib
import io
import os
from typing import Any, Callable, ContextManager, IO, Iterable, Optional, Set, Union


class VirtualFiles:
  """
  Represents a collection of files that are represented virtually either by static
  text or a function that can render the contents into a file-like object. The files
  can then be written to disk in one go.
  """

  def __init__(self):
    self._files = []

  def update(self, other: 'VirtualFiles', prefix: Optional[str] = None) -> None:
    for item in other._files:
      item = dict(item)
      if prefix:
        item['filename'] = os.path.join(prefix, item['filename'])
      self._files.append(item)

  def add_static(self, filename: str, content: Union[str, bytes]) -> None:
    def _write(fp):
      fp.write(content)
    self.add_dynamic(filename, _write, text=isinstance(content, str))

  def add_dynamic(
    self,
    filename: str,
    render_func: Callable,
    *args: Any,
    text: bool = True,
    inplace: bool = False,
  ) -> None:
    self._files.append({
      'filename': filename,
      'render_func': render_func,
      'args': args,
      'text': text,
      'inplace': inplace,
    })

  def write_all(
    self,
    parent_directory: str = None,
    on_write: Callable = None,
    on_skip: Callable = None,
    overwrite: bool = False,
    create_directories: bool = True,
    dry: bool = False,
    open_func: Callable[[str, str], ContextManager[IO]] = None,

  ) -> None:
    """
    Writes all files to disk. Relative files will be written relative to the
    *parent_directory*.
    """

    if open_func is None:
      open_func = open

    for file_, filename in zip(self._files, self.abspaths(parent_directory)):
      exists = os.path.isfile(filename)
      if exists and not overwrite:
        if on_skip:
          on_skip(filename)
        continue
      if on_write:
        on_write(filename)
      if not dry:
        mode = '' if file_['text'] else 'b'
        if create_directories:
          os.makedirs(os.path.dirname(filename) or '.', exist_ok=True)
        if file_['inplace']:
          with open_func(filename, 'w' + mode) as dst:
            if exists:
              # TODO: This needs to use an atomic file actually..
              with open(filename, 'r' + mode) as src:
                file_['render_func'](dst, src, *file_['args'])
            else:
              file_['render_func'](dst, None, *file_['args'])
        else:
          with open_func(filename, 'w' + mode) as dst:
            file_['render_func'](dst, *file_['args'])

  def abspaths(self, parent_directory: str = None) -> Iterable[str]:
    """
    Returns all paths in this virtual fileset joined with *parent_directory*.
    """

    for file_ in self._files:
      yield os.path.normpath(os.path.join(parent_directory or '.', file_['filename']))

  def get_modified_files(self, parent_directory: str) -> Set[str]:
    """
    Returns a set of the files that would be modified by writing the virtual files to disk.
    """

    modified_files = set()

    @contextlib.contextmanager
    def opener(filename, mode):
      fp = io.BytesIO() if 'b' in mode else io.StringIO()
      yield fp
      if not os.path.isfile(filename):
        modified_files.add(filename)
      else:
        with open(filename, mode.replace('w', 'r')) as src:
          if fp.getvalue() != src.read():
            modified_files.add(filename)

    self.write_all(parent_directory, open_func=opener, overwrite=True, dry=False)
    return {os.path.relpath(f, parent_directory) for f in modified_files}



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides functions to render a list of changelogs. Always only supports the newest
changelog version.
"""

from .manager import Changelog
from nr.stream import Stream  # type: ignore
from termcolor import colored
from typing import List, TextIO
import re
import shutil
import textwrap


def _group_entries_by_component(entries):
  key = lambda x: x.component
  return list(Stream.sortby(entries, key).groupby(key, collect=list))


def _terminal(fp: TextIO, changelogs: List[Changelog]) -> None:
  """
  Renders a #Changelog for the terminal to *fp*.
  """

  def _md_term_stylize(text: str) -> str:
    def _code(m):
      return colored(m.group(1), 'cyan')
    def _issue_ref(m):
      return colored(m.group(0), 'yellow', attrs=['bold'])
    text = re.sub(r'`([^`]+)`', _code, text)
    text = re.sub(r'#\d+', _issue_ref, text)
    return text

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.fixes:
      return None
    return '(' + ', '.join(colored(_fmt_issue(i), 'yellow', attrs=['underline']) for i in entry.fixes) + ')'

  def _fmt_types(entry):
    return colored(entry.type_.name, attrs=['bold'])

  if hasattr(shutil, 'get_terminal_size'):
    width = shutil.get_terminal_size((80, 23))[0]
  else:
    width = 80

  # Explode entries by component.
  for changelog in changelogs:
    fp.write(colored(changelog.version or 'Unreleased', 'blue', attrs=['bold', 'underline']))
    fp.write(' ({})\n'.format(changelog.data.release_date or 'no release date'))
    for component, entries in _group_entries_by_component(changelog.entries):
      maxw = max(map(lambda x: len(x.type_.name), entries))
      fp.write('  ' + colored(component or 'No Component', 'red', attrs=['bold', 'underline']) + '\n')
      for entry in entries:
        lines = textwrap.wrap(entry.description, width - (maxw + 6))
        suffix_fmt = ' '.join(filter(bool, (_fmt_issues(entry),)))
        lines[-1] += ' ' + suffix_fmt
        delta = maxw - len(entry.type_.name)
        fp.write('    {} {}\n'.format(colored((_fmt_types(entry) + ':') + ' ' * delta, attrs=['bold']), _md_term_stylize(lines[0])))
        for line in lines[1:]:
          fp.write('    {}{}\n'.format(' ' * (maxw+2), _md_term_stylize(line)))
    fp.write('\n')


def _markdown(fp: TextIO, changelogs: List[Changelog]) -> None:

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.fixes:
      return None
    return '(' + ', '.join(_fmt_issue(i) for i in entry.fixes) + ')'

  for changelog in changelogs:
    fp.write('## {}'.format(changelog.version or 'Unreleased'))
    fp.write(' ({})\n\n'.format(changelog.data.release_date or 'no release date'))
    for component, entries in _group_entries_by_component(changelog.entries):
      fp.write('* __{}__\n'.format(component))
      for entry in entries:
        description ='**' + entry.type_.name + '**: ' + entry.description
        if entry.fixes:
          description += ' ' + _fmt_issues(entry)
        lines = textwrap.wrap(description, 80)
        fp.write('    * {}\n'.format(lines[0]))
        for line in lines[1:]:
          fp.write('      {}\n'.format(line))
    fp.write('\n')


renderers = {
  'terminal': _terminal,
  'markdown': _markdown,
}


def render(fp: TextIO, format: str, changelogs: List[Changelog]) -> None:
  changelogs = list(changelogs)
  unreleased = next((x for x in changelogs if not x.version), None)
  if unreleased:
    changelogs.remove(unreleased)
  changelogs.sort(key=lambda x: x.version, reverse=True)
  if unreleased:
    changelogs.insert(0, unreleased)

  renderers[format](fp, changelogs)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
import time

import click
from termcolor import colored

from shut.commands.pkg.test import test_package, print_test_run
from shut.model.monorepo import MonorepoModel
from shut.test.base import TestStatus
from . import mono, project


@mono.command()
@click.option('--isolate/--no-isolate', default=False,
  help='Isolate all test runs in virtual environments. This greatly increases the duration '
       'for tests to run as the environment needs to be created and packages installed first, '
       'but it ensures that the unit tests work for a vanilla installation (default: false)')
@click.option('--keep-test-env', is_flag=True,
  help='Do not delete the virtual environment created when testing with --isolate.')
@click.option('--capture/--no-capture', default=True,
  help='Capture the output of the underlying testing framework. If set to false, the output '
       'will be routed to stderr (default: true)')
@click.option('--only', help='Comma-separated list of packages to test.')
def test(isolate: bool, keep_test_env: bool, capture: bool, only: str) -> None:
  """
  Run unit tests for all packages in the mono repository.
  """

  monorepo = project.load_or_exit(expect=MonorepoModel)

  if only:
    packages = []
    package_map = {p.name: p for p in project.packages}
    for package_name in only.split(','):
      if package_name not in package_map:
        sys.exit(f'error: package "{package_name}" does not exist')
      package = package_map[package_name]
      if not package.test_driver:
        sys.exit(f'error: package "{package_name}" has no test driver configured')
      packages.append(package)
  else:
    packages = list(filter(lambda p: p.test_driver, project.packages))

  packages = sorted(packages, key=lambda p: p.name)

  print(f'Going to test {len(packages)} package(s):')
  for package in packages:
    print(f'  {colored(package.name, "yellow")}')
  print()

  exit_code = 0
  all_tests = []
  all_errors = []
  package_statuses = []
  tstart = time.perf_counter()

  for i, package in enumerate(packages):
    if i > 0:
      print()
    print(f'Testing package {colored(package.name, "yellow", attrs=["bold"])}:')
    print()
    test_run = test_package(package, isolate, keep_test_env, capture)
    all_tests += test_run.tests
    all_errors += test_run.errors
    print_test_run(test_run)
    if test_run.status != TestStatus.PASSED:
      exit_code = 1
    package_statuses.append((package, test_run.status))

  n_passed = sum(1 for t in all_tests if t.status == TestStatus.PASSED)
  n_skipped = sum(1 for t in all_tests if t.status == TestStatus.SKIPPED)
  duration = time.perf_counter() - tstart

  print()
  print(colored('Monorepo summary:', attrs=['bold', 'underline']))
  print()
  print(f'Ran {len(all_tests)} test(s) in {duration:.3f}s ({n_passed} passed, '
        f'{n_skipped} skipped, {len(all_tests) - n_passed} failed, {len(all_errors)} error(s)). '
        f'{"PASSED" if exit_code == 0 else "FAILED"}')
  print()
  for package, status in package_statuses:
    color = {TestStatus.PASSED: 'green', TestStatus.SKIPPED: 'yellow', TestStatus.FAILED: 'red', TestStatus.ERROR: 'red'}[status]
    print(f'  {colored(package.name, color, attrs=["bold"])} {status.name}')

  sys.exit(exit_code)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import datetime
import json
import os
import subprocess as sp
import sys
from typing import List, Optional, Tuple, TYPE_CHECKING

from databind.core import datamodel, field
from nr.parsing.date import timezone  # type: ignore

from shut.model.requirements import Requirement
from .base import (BaseTestDriver, Runtime, StackTrace, TestCase, TestCrashReport,
  TestEnvironment, TestError, TestRun, TestStatus)

if TYPE_CHECKING:
  from shut.model.package import PackageModel


def load_report_file(report_file: str) -> TestRun:
  """
  Loads a Pytest-json-report file into a #TestRun.
  """

  with open(report_file) as fp:
    raw = json.load(fp)

  started_time = datetime.datetime.fromtimestamp(raw['created']).replace(tzinfo=timezone.local).astimezone(timezone.utc)
  duration = raw['duration']
  status = TestStatus.PASSED if raw['exitcode'] == 0 else TestStatus.FAILED
  environment = TestEnvironment(raw['environment']['Python'], raw['environment']['Platform'])
  errors: List[TestError] = []
  tests: List[TestCase] = []

  # Map the nodes to the line and function that they are defined in.
  testid_to_source: Dict[str, Tuple[str, int]] = {}
  for node in raw['collectors']:
    if node['nodeid'] and node['result']:
      for result in node['result']:
        if 'lineno' in result:
          testid_to_source[result['nodeid']] = (node['nodeid'], result['lineno'])
    if node['outcome'] != 'passed':
      errors.append(TestError(node['nodeid'], node['longrepr']))

  # Collect the test results.
  for test in raw['tests']:
    failed_stage = next((test[k] for k in ('setup', 'call', 'teardown')
      if k in test and test[k]['outcome'] == 'failed'), None)
    if failed_stage:
      crash = TestCrashReport(
        filename=failed_stage['crash']['path'],
        lineno=failed_stage['crash']['lineno'],
        message=failed_stage['crash']['message'],
        traceback=[StackTrace(trace['path'], trace['lineno'], trace['message']) for trace in failed_stage['traceback']],
        longrepr=failed_stage['longrepr'],
      )
      stdout = failed_stage.get('stdout')
    else:
      crash = None
      stdout = None
    test_status = {'passed': TestStatus.PASSED, 'failed': TestStatus.FAILED, 'skipped': TestStatus.SKIPPED}[test['outcome']]
    tests.append(TestCase(
      name=test['nodeid'],
      duration=sum(test[k]['duration'] for k in ('setup', 'call', 'teardown') if k in test),
      filename=testid_to_source[test['nodeid']][0],
      lineno=testid_to_source[test['nodeid']][1],
      status=test_status,
      crash=crash,
      stdout=stdout,
    ))

  return TestRun(started_time, duration, status, environment, tests, errors)


@datamodel
class PytestDriver(BaseTestDriver):
  """
  A driver for running unit tests using [Pytest][1].

  [1]: https://docs.pytest.org/en/latest/
  """

  directory: Optional[str] = None
  args: List[str] = field(default_factory=lambda: ['-vv'])
  report_file: str = field(default='.pytest-report.json', altname='report-file')

  # BaseTestDriver

  def test_package(self, package: 'PackageModel', runtime: Runtime, capture: bool) -> TestRun:
    test_dir = os.path.join(package.source_directory, self.directory) if self.directory else package.source_directory
    test_dir = os.path.join(package.get_directory(), test_dir)
    command = runtime.python + ['-m', 'pytest', test_dir]
    command += ['--json-report', '--json-report-file', self.report_file]
    command += self.args

    if os.path.isfile(self.report_file):
      os.remove(self.report_file)

    started_time = datetime.datetime.now(timezone.utc)
    proc = sp.Popen(command, stdout=sp.PIPE if capture else sys.stderr,
      stderr=sp.STDOUT if capture else None)
    output, _ = proc.communicate()
    result = proc.wait()

    if os.path.isfile(self.report_file):
      test_run = load_report_file(self.report_file)
      if result != 0:
        test_run.status = TestStatus.FAILED
      os.remove(self.report_file)
    else:
      test_run = TestRun(
        started=started_time,
        duration=(datetime.datetime.now(timezone.utc) - started_time).total_seconds(),
        status=TestStatus.ERROR,
        environment=runtime.get_environment(),
        tests=[],
        errors=[TestError(None, output.decode())],
      )

    return test_run

  def get_test_requirements(self) -> List[Requirement]:
    return [
      Requirement.parse('pytest'),
      Requirement.parse('pytest-json-report ^1.2.1'),
    ]



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys

from nr.utils.git import Git  # type: ignore
from termcolor import colored

from shut.model import AbstractProjectModel, MonorepoModel, Project


def get_commits_since_last_tag(subject: AbstractProjectModel):
  tag = subject.get_tag(subject.get_version())
  ref = Git().rev_parse(tag)
  if not ref:
    return tag, None
  else:
    return tag, len(Git().rev_list(tag + '..HEAD', subject.get_directory()))


def print_status(project: Project) -> None:
  """
  The latest version is taken from the current version number in the package
  configuration file. Git is then queried for the tag and the commit distance
  to the current revision.
  """

  assert project.subject, "No subject"

  if isinstance(project.subject, MonorepoModel):
    monorepo_dir = project.subject.get_directory()
    if not project.packages:
      sys.exit('error: monorepo has no packages')
    items = sorted(project.packages, key=lambda x: x.name)
    names = [os.path.normpath(os.path.relpath(x.get_directory(), monorepo_dir)) for x in items]
  else:
    items = [project.subject]
    names = [project.subject.get_name()]

  width = max(map(len, names)) if names else 0

  for item, name in zip(items, names):
    tag, num_commits = get_commits_since_last_tag(item)
    if num_commits is None:
      item_info = colored('tag "{}" not found'.format(tag), 'red')
    elif num_commits == 0:
      item_info = colored('no commits', 'green') + ' since "{}"'.format(tag)
    else:
      item_info = colored('{} commit(s)'.format(num_commits), 'yellow') + ' since "{}"'.format(tag)
    print('{}: {}'.format(name.rjust(width), item_info))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import shutil
import sys
from typing import Iterable, List, Optional

from shut.model import PackageModel
from shut.model.target import TargetId
from shut.utils.io.sp import subprocess_trimmed_call
from .core import Builder, BuilderProvider, register_builder_provider


class SetuptoolsBuilder(Builder):
  """
  Internal. Implements building a Python package.
  """

  _FORMATS_MAP = {
    'zip': '.zip',
    'gztar': '.tar.gz',
    'bztar': '.tar.bz2',
    'ztar': '.tar.Z',
    'tar': '.tar'
  }

  def __init__(
    self,
    id_: TargetId,
    description: str,
    output_files: List[str],
    package_directory: str,
    build_type: str,
    args: List[str],
  ) -> None:
    self._id = id_
    self.description = description
    self.output_files = output_files
    self.package_directory = package_directory
    self.build_type = build_type
    self.args = args

  def __repr__(self):
    return f'SetuptoolsBuilder(id={self.id!r}, build_type={self.build_type!r}, args={self.args!r})'

  @classmethod
  def wheel(
    cls,
    id_: TargetId,
    description: str,
    package: PackageModel,
  ) -> 'SetuptoolsBuilder':
    py = 'py2.py3' if package.is_universal() else ('py' + sys.version[0])
    filename = f'{package.name.replace("-", "_")}-{package.version}-{py}-none-any.whl'
    return cls(id_, description, [filename], package.get_directory(), 'bdist_wheel', [])

  @classmethod
  def sdist(
    cls,
    id_: TargetId,
    description: str,
    formats: List[str],
    package: PackageModel,
  ) -> 'SetuptoolsBuilder':
    assert formats
    return cls(
      id_,
      description,
      [f'{package.name}-{package.version}{cls._FORMATS_MAP[f]}' for f in formats],
      package.get_directory(),
      'sdist',
      ['--format', ','.join(formats)],
    )

  # Builder Overrides

  def get_description(self) -> Optional[str]:
    return self.description

  def get_outputs(self) -> Iterable[str]:
    return self.output_files

  def build(self, build_directory: str, verbose: bool) -> bool:
    # TODO: Can we change the distribution output directory with an option?
    python = os.getenv('PYTHON', sys.executable)
    dist_directory = os.path.join(self.package_directory, 'dist')
    dist_exists = os.path.exists(dist_directory)
    command = [python, 'setup.py', self.build_type] + self.args

    res = subprocess_trimmed_call(command, cwd=self.package_directory)
    if res != 0:
      return False

    # Make sure the files end up in the correct directory.
    for filename in self.output_files:
      src = next(filter(os.path.isfile, [
        os.path.join(dist_directory, filename),
        os.path.join(dist_directory, filename.lower())]), None)
      if not src:
        raise RuntimeError('{} not produced by setup.py {}'.format(filename, self.build_type))
      dst = os.path.join(build_directory, filename)
      if src != dst:
        if os.path.isfile(dst):
          os.remove(dst)
        os.rename(src, dst)

    # Cleanup after yourself.
    if not dist_exists:
      shutil.rmtree(dist_directory)

    return True

  # Target Overrides

  @property
  def id(self) -> TargetId:
    return self._id


class SetuptoolsBuilderProvider(BuilderProvider[PackageModel]):

  # BuilderProvider Overrides

  def get_builders(self, package: PackageModel) -> Iterable[Builder]:
    yield SetuptoolsBuilder.sdist(
      TargetId('setuptools', 'sdist'),
      'Build a source distribute',
      ['gztar'],
      package,
    )

    if package.wheel:
      yield SetuptoolsBuilder.wheel(
        TargetId('setuptools', 'wheel'),
        'Build a Python wheel.',
        package,
      )


register_builder_provider(PackageModel, SetuptoolsBuilderProvider)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import shlex
import subprocess
import sys

import nr.fs  # type: ignore

__all__ = ['editor_open', 'edit_text']


def editor_open(filename: str):
  editor = shlex.split(os.getenv('EDITOR', 'vim'))
  return subprocess.call(editor + [filename])


def edit_text(text: str) -> str:
  """
  Opens an editor for the user to modify *text*.
  """

  with nr.fs.tempfile('.yml', dir=os.getcwd(), text=True) as fp:
    fp.write(text)
    fp.close()
    res = editor_open(fp.name)
    if res != 0:
      sys.exit(res)
    with open(fp.name) as src:
      return src.read()



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import subprocess
from typing import Any, Dict, Optional

import jinja2
import nr.fs  # type: ignore
from termcolor import colored

from shut.model.author import Author
from shut.utils.external.license import get_license_metadata, wrap_license_text
from shut.utils.io.virtual import VirtualFiles

GITIGNORE_TEMPLATE = '''
.venv*/
dist/
build/
*.py[cod]
*.egg-info
*.egg
'''.lstrip()

README_TEMPLATE = '''
# {{project_name}}

---

<p align="center">Copyright &copy; {{year}} {{author.name}}</p>
'''.lstrip()


def load_author_from_git() -> Optional[str]:
  """
  Returns a string formatted as "name <mail>" from the Git `user.name` and `user.email`
  configuration values. Returns `None` if Git is not configured.
  """

  try:
    name = subprocess.getoutput('git config user.name')
    email = subprocess.getoutput('git config user.email')
  except FileNotFoundError:
    return None
  if not name and not email:
    return None
  return Author(name, email)


def get_license_file_text(license: str, template_vars: Dict[str, Any]) -> str:
  license_text = 'Copyright (c) {year} {author.name}\n\n'.format(**template_vars)
  license_text += wrap_license_text(get_license_metadata(license)['license_text'])
  return license_text


def render_template(fp, template_string, template_vars):
  for data in jinja2.Template(template_string).stream(**template_vars):
    fp.write(data)
  fp.write('\n')


def write_files(
  files: VirtualFiles,
  target_directory: str,
  force: bool = False,
  dry: bool = False,
  indent: int = 0,
):
  def _rel(fn: str) -> str:
    path = os.path.relpath(fn)
    if nr.fs.issub(path):
      return path
    return fn
  str_indent = '  ' * indent
  files.write_all(
    target_directory,
    on_write=lambda fn: print(str_indent + colored('write ' + _rel(fn), 'cyan')),
    on_skip=lambda fn: print(str_indent + colored('skip ' + _rel(fn), 'yellow')),
    overwrite=force,
    dry=dry,
  )



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from shut.checkers import get_checks
from shut.commands import project
from shut.commands.commons.checks import print_checks_all, get_checks_status
from shut.commands.mono import mono
from shut.model import MonorepoModel

import click
import logging
import sys
import time

logger = logging.getLogger(__name__)


def check_monorepo(
  monorepo: MonorepoModel,
  warnings_as_errors: bool = False,
  skip_positive_checks: bool = False,
  print_stats: bool = True,
  use_stderr: bool = False,
) -> int:
  start_time = time.perf_counter()
  checks = sorted(get_checks(monorepo), key=lambda c: c.name)
  seconds = time.perf_counter() - start_time
  print_checks_all(monorepo.name, checks, seconds,
    skip_positive_checks=skip_positive_checks, print_stats=print_stats,
    use_stderr=use_stderr)
  return get_checks_status(checks, warnings_as_errors)


@mono.command()
@click.option('-w', '--warnings-as-errors', is_flag=True)
def checks(warnings_as_errors):
  """
  Sanity-check the package configuration and package files. Which checks are performed
  depends on the features that are enabled in the package configuration. Usually that
  will at least include the "setuptools" feature which will perform basic sanity checks
  on the package configuration and entrypoint definition.
  """

  monorepo = project.load_or_exit(expect=MonorepoModel)
  sys.exit(check_monorepo(monorepo, warnings_as_errors))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
from typing import Iterable, Generic, Type, TypeVar

from databind.core import datamodel

from shut.utils.io.virtual import VirtualFiles
from shut.utils.type_registry import TypeRegistry

T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)

__all__ = [
  'VersionRef',
  'Renderer',
  'register_renderer',
  'get_files',
  'get_version_refs',
]


@datamodel
class VersionRef:
  filename: str
  start: int
  end: int
  value: str


class Renderer(Generic[T_co], metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_files(self, files: VirtualFiles, obj: T) -> None:
    pass

  def get_version_refs(self, obj: T) -> Iterable[VersionRef]:
    return; yield


registry = TypeRegistry[Renderer[T_co]]()


def register_renderer(t: Type[T_co], renderer: Type[Renderer[T_co]]) -> None:
  """
  Register the *renderer* implementation to run when creating files for *t*.
  """

  registry.put(t, renderer)


def get_files(obj: T) -> VirtualFiles:
  """
  Gets all the files from the renderers registered to the type of *obj*.
  """

  files = VirtualFiles()
  for renderer in registry.for_type(type(obj)):
    renderer().get_files(files, obj)

  return files


def get_version_refs(obj: T) -> Iterable[VersionRef]:
  """
  Gets all version refs returned by registered for type *T_co*.
  """

  for renderer in registry.for_type(type(obj)):
    yield from renderer().get_version_refs(obj)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.


import datetime
import os
from typing import Iterable, Optional, Union

from databind.core import datamodel, field
from databind.json import from_json, to_json
import yaml

from shut.model import registry
from shut.model.version import Version
from . import v1, v2, v3

AllChangelogTypes = Union[v3.Changelog, v2.Changelog, v1.Changelog]


@datamodel
class Changelog:
  """
  Represents a changelog on disk.
  """

  filename: Optional[str] = field(derived=True, default=None)
  version: Optional[Version]
  data: v3.Changelog = field(default_factory=lambda: v3.Changelog(changes=[]))

  @property
  def entries(self):
    return self.data.changes

  def exists(self) -> bool:
    " Returns #True if the changelog file exists. "

    return os.path.isfile(self.filename)

  def load(self) -> None:
    " Loads the data from the file of this changelog. "

    with open(self.filename) as fp:
      raw_data = yaml.safe_load(fp)

    data = from_json(AllChangelogTypes, raw_data, registry=registry)
    if not isinstance(data, v3.Changelog):
      data = v3.Changelog.migrate(data)

    self.data = data

  def save(self, create_directory: bool = False) -> None:
    " Saves the changelog. It will always save the changelog in the newest supported format. "

    if create_directory:
      os.makedirs(os.path.dirname(self.filename), exist_ok=True)
    data = to_json(self.data, v3.Changelog)
    with open(self.filename, 'w') as fp:
      yaml.safe_dump(data, fp, sort_keys=False)

  def set_release_date(self, date: datetime.date) -> None:
    self.data.release_date = date

  def add_entry(self, entry) -> None:
    assert isinstance(entry, v3.Changelog.Entry), type(entry)
    self.data.changes.append(entry)


class ChangelogManager:

  def __init__(self, directory: str) -> None:
    self.directory = directory
    self._cache = {}

  def _get(self, name: str, version: Optional[str]) -> Changelog:
    key = (name, str(version))
    if key in self._cache:
      return self._cache[key]
    changelog = Changelog(os.path.join(self.directory, name), version)
    if os.path.isfile(changelog.filename):
      changelog.load()
    self._cache[key] = changelog
    return changelog

  @property
  def unreleased(self) -> Changelog:
    return self._get('_unreleased.yml', None)

  def version(self, version: Version) -> Changelog:
    return self._get(str(version) + '.yml', version)

  def release(self, version: Version) -> Changelog:
    """
    Renames the unreleased changelog to the file name for the specified *version*.
    """

    unreleased = self.unreleased
    unreleased.data.release_date = datetime.date.today()
    unreleased.save()

    os.rename(unreleased.filename, self.version(version).filename)
    self._cache.clear()

    return self.version(version)

  def all(self) -> Iterable[Changelog]:
    """
    Yields all changelogs.
    """

    for name in os.listdir(self.directory):
      if not name.endswith('.yml'):
        continue
      if name == '_unreleased.yml':
        yield self.unreleased
      else:
        version = Version(name[:-4])
        yield self.version(version)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
from fnmatch import fnmatch

from databind.core import datamodel


@datamodel(frozen=True)
class TargetId:
  """
  Represents the ID of a target that can be selected on the command line. We use the
  "target" concept for builds and publishers. A target ID may contain wildcard characters
  in which case it can be used to match multiple targets with #fnmatch().
  """

  scope: str
  name: str

  def __str__(self):
    return f'{self.scope}:{self.name}'

  @classmethod
  def parse(cls, s: str, allow_scope_only: bool = False) -> 'TargetId':
    parts = s.split(':')
    if allow_scope_only and len(parts) == 1:
      parts = [parts[0], '*']
    return cls(*parts)

  def match(self, other_id: 'TargetId', allow_match_name: bool = False) -> bool:
    if fnmatch(other_id.scope, self.scope) and fnmatch(other_id.name, self.name):
      return True
    if allow_match_name and self.name == '*':
      return self.scope == other_id.name
    return False


class Target(metaclass=abc.ABCMeta):

  @abc.abstractproperty
  def id(self) -> TargetId:
    pass



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
from typing import Generic, Iterable, List, Type, TypeVar

from nr.stream import concat  # type: ignore

from shut.model import AbstractProjectModel
from shut.model.target import Target
from shut.utils.type_registry import TypeRegistry

T = TypeVar('T')

__all__ = [
  'Builder',
  'BuilderProvider',
  'register_builder_provider',
  'get_builders',
]


class Builder(Target, metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_outputs(self) -> Iterable[str]:
    """
    Returns a list of the output files produced by this builder.
    """

  @abc.abstractmethod
  def build(self, build_directory: str, verbose: bool) -> bool:
    """
    Run the build. Output from subprocesses should be captured unless *verbose* is enabled.
    """


class BuilderProvider(Generic[T], metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_builders(self, obj: T) -> List[Builder]:
    pass


registry = TypeRegistry[AbstractProjectModel]()


def register_builder_provider(type_: Type[T], provider_class: Type[BuilderProvider[T]]) -> None:
  registry.put(type_, provider_class)


def get_builders(obj: T) -> Iterable[Builder]:
  return concat(provider().get_builders(obj) for provider in registry.for_type(type(obj)))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import re
from typing import Iterable

from shut.model import AbstractProjectModel, MonorepoModel
from shut.utils.io.virtual import VirtualFiles
from .core import Renderer, get_version_refs, register_renderer, VersionRef


class GenericRenderer(Renderer[AbstractProjectModel]):

  # Renderer[AbstractProjectModel] Overrides

  def get_files(self, files: VirtualFiles, obj: AbstractProjectModel) -> None:
    pass

  def get_version_refs(self, obj: AbstractProjectModel) -> Iterable[VersionRef]:
    assert obj.filename
    assert obj.project

    # Return a reference to the version number in the package or monorepo model.
    regex = '^\s*version\s*:\s*[\'"]?(.*?)[\'"]?\s*(#.*)?$'
    with open(obj.filename) as fp:
      match = re.search(regex, fp.read(), re.S | re.M)
      if match:
        yield VersionRef(obj.filename, match.start(1), match.end(1), match.group(1))

    if isinstance(obj, MonorepoModel) and obj.release.single_version:
      for package in obj.project.packages:
        yield from get_version_refs(package)


register_renderer(AbstractProjectModel, GenericRenderer)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from databind.core import datamodel
import re


@datamodel
class Author:
  """
  Represents information about an author. Can be deserialized from a string of the form
  `Name <user@domain.name>`
  """

  AUTHOR_EMAIL_REGEX = re.compile(r'([^<]+)<([^>]+)>')
  name: str
  email: str

  @classmethod
  def parse(cls, string: str) -> 'Author':
    match = Author.AUTHOR_EMAIL_REGEX.match(string)
    if not match:
      raise ValueError('not a valid author string: {!r}'.format(string))
    name = match.group(1).strip()
    email = match.group(2).strip()
    return cls(name, email)

  def __str__(self):
    return '{} <{}>'.format(self.name, self.email)

  @classmethod
  def databind_json_load(cls, value, context):
    if isinstance(value, str):
      return cls.parse(value)
    return NotImplemented

  def databind_json_dump(self, context):
    return str(self)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
A tiny client for the [DejaCode license library][1].

  [1]: https://enterprise.dejacode.com/licenses/
"""

import bs4
import re
import requests
import textwrap

BASE_URL = 'https://enterprise.dejacode.com/licenses/public/{}/'


def _get_table_value_by_key(soup, key):
  regex = re.compile('\s*' + re.escape(key) + '\s*')
  item = soup.find('span', text=regex)
  if item is None:
    raise ValueError('<span/> for {!r} not found'.format(key))
  value = item.parent.findNext('dd').find('pre').text
  if value == '\xa0':
    value = ''
  return value or None


def _get_license_text(soup):
  tab = soup.find(id='tab_license-text')
  if tab is None:
    raise ValueError('#tab_license-text not found')
  pre = soup.find('div', {'class': 'clipboard'}).find('pre')
  return pre.text


def get_license_metadata(license_name):
  """ Retrives the HTML metadata page for the specified license from the
  DejaCode website and extracts information such as the name, category,
  license type, standard notice and license text. """

  url = BASE_URL.format(license_name.replace(' ', '-').lower())
  response = requests.get(url)
  response.raise_for_status()
  html = response.text
  soup = bs4.BeautifulSoup(html, 'html.parser')

  extract_keys = ['Key', 'Name', 'Short Name', 'Category', 'License type',
    'License profile', 'License style', 'Owner', 'SPDX short identifier',
    'Keywords', 'Standard notice', 'Special obligations', 'Publication year',
    'URN', 'Dataspace', 'Homepage URL', 'Text URLs', 'OSI URL', 'FAQ URL',
    'Guidance URL', 'Other URLs']

  data = {}
  for key in extract_keys:
    data[key.replace(' ', '_').lower()] = _get_table_value_by_key(soup, key)
  data['publication_year'] = int(data['publication_year'])
  if data['standard_notice']:
    data['standard_notice'] = textwrap.dedent(data['standard_notice'])
  data['license_text'] = _get_license_text(soup)

  return data


def wrap_license_text(license_text, width=79):
  lines = []
  for line in license_text.split('\n'):
    line = line.split(' ')
    length = sum(map(len, line)) + len(line) - 1
    if length > width:
      words = []
      length = -1
      for word in line:
        if length + 1 + len(word) >= width:
          lines.append(' '.join(words))
          words = []
          length = -1
        else:
          words.append(word)
          length += len(word) + 1
      if words:
        lines.append(' '.join(words))
    else:
      lines.append(' '.join(line))
  return '\n'.join(lines)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import shlex
import subprocess as sp
import sys
import typing as t
from typing import cast, List, Optional, Set

import click
import networkx as nx

from shut.model.package import PackageModel
from shut.model.monorepo import InterdependencyRef
from shut.model.requirements import RequirementsList, VendoredRequirement
from . import pkg
from .. import project


def collect_requirement_args(
  package: bool,
  develop: bool,
  inter_deps: bool,
  extra: Optional[Set[str]],
) -> List[str]:
  reqs = RequirementsList()

  # Pip does not understand "test" as an extra and does not have an option to
  # install test requirements.
  if extra and 'test' in extra:
    reqs += package.test_requirements

  reqs.append(VendoredRequirement(VendoredRequirement.Type.Path, package.get_directory()))
  reqs += package.requirements.vendored_reqs()

  if project.monorepo and inter_deps:
    # TODO(NiklasRosenstein): get_inter_dependencies_for() does not currently differentiate
    #   between normal, test and extra requirements.

    # In the case of transitive inter-dependencies, we need to make sure to install them
    # in the right order. For example in the command-line `pip install -e a -e b -e c`, where
    # package `a` depends on package `b`, Pip is not able to tell that `b` is supplied on the
    # command-line as well and will instead try to resolve it from PyPI.
    project_packages: t.Dict[str, PackageModel] = {p.name: p for p in project.packages}
    direct_inter_dependencies: t.Dict[str, InterdependencyRef]  = {ref.package_name: ref
      for ref in project.monorepo.get_inter_dependencies_for(package)}
    graph: nx.DiGraph = project.monorepo.get_inter_dependencies_graph().subgraph(
      direct_inter_dependencies)

    insert_index = 0
    for dep_name in nx.algorithms.topological_sort(graph):
      dep = project_packages[dep_name]
      ref = direct_inter_dependencies[dep_name]
      if not ref.version_selector.matches(dep.version):
        print('note: skipping inter-dependency on package "{}" because the version selector '
              '{} does not match the present version {}'
              .format(dep.name, ref.version_selector, dep.version), file=sys.stderr)
      else:
        reqs.insert(insert_index, VendoredRequirement(VendoredRequirement.Type.Path, dep.get_directory()))
        insert_index += 1

  return reqs.get_pip_args(package.get_directory(), develop)


def run_install(
  pip: Optional[str],
  args: List[str],
  upgrade: bool,
  quiet: bool,
  dry: bool,
) -> None:

  pip_bin = shlex.split(os.getenv('PIP', pip or 'python -m pip'))
  command = pip_bin + ['install'] + args
  if upgrade:
    command.append('--upgrade')
  if quiet:
    command.append('--quiet')

  if dry:
    print(' '.join(map(shlex.quote, command)))
  else:
    sys.exit(sp.call(command))


def split_extras(extras: str) -> Set[str]:
  result = set(map(str.strip, extras.split(',')))
  result.discard('')
  return result


@pkg.command()
@click.option('--develop/--no-develop', default=True,
  help='Install in develop mode (default: true)')
@click.option('--inter-deps/--no-inter-deps', default=True,
  help='Install package inter dependencies from inside the same monorepo (default: true)')
@click.option('--extra', type=split_extras, help='Specify one or more extras to install.')
@click.option('-U', '--upgrade', is_flag=True, help='Upgrade all packages (forwarded to pip install).')
@click.option('-q', '--quiet', is_flag=True, help='Quiet install')
@click.option('-v', '--verbose', is_flag=True, help='Verbose install')
@click.option('--pip', help='Override the command to run Pip. Defaults to "python -m pip" or the PIP variable.')
@click.option('--pip-args', help='Additional arguments to pass to Pip.')
@click.option('--dry', is_flag=True, help='Print the Pip command to stdout instead of running it.')
@click.option('--pipx', is_flag=True, help='Install using Pipx.')
def install(develop, inter_deps, extra, upgrade, quiet, verbose, pip, pip_args, dry, pipx):
  """
  Install the package using `python -m pip`. If the package is part of a mono repository,
  inter-dependencies will be installed from the mono repsitory rather than from PyPI.

  The command used to invoke Pip can be overwritten using the `PIP` environment variable.
  """

  if not pip and pipx:
    pip = 'pipx'

  package = cast(PackageModel, project.load_or_exit(expect=PackageModel))
  args = collect_requirement_args(package, develop, inter_deps, extra)
  if pipx:
    args += ['--pip-args', ' '.join(map(shlex.quote, package.install.get_pip_args()))]
  else:
    args += package.install.get_pip_args()
  if verbose:
    pip_args.append('--verbose')
  if pip_args:
    args += shlex.split(pip_args)
  run_install(pip, args, upgrade, quiet, dry)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import builtins
import sys
from functools import partial
from typing import List

import termcolor

from shut.checkers import Check, CheckStatus


def print_checks(
  checks: List[Check],
  emojis: bool = True,
  colors: bool = True,
  prefix: str = '',
  skip_positive_checks: bool = False,
  use_stderr: bool = False,
) -> None:
  """
  Prints the list of *checks* to the terminal.
  """

  emoji_chars = {
    CheckStatus.PASSED: '✔️ ',
    CheckStatus.WARNING: '⚠️ ',
    CheckStatus.ERROR: '❌'}

  color_names = {
    CheckStatus.PASSED: 'green',
    CheckStatus.WARNING: 'yellow',
    CheckStatus.ERROR: 'red'}

  if colors:
    colored = termcolor.colored
  else:
    def colored(s, *a, **kw):
      return str(s)

  out = sys.stderr if use_stderr else sys.stdout
  print = partial(builtins.print, file=out)

  num_printed = 0
  for check in checks:
    if skip_positive_checks and check.result.status == CheckStatus.PASSED:
      continue
    if num_printed == 0:
      print()
    num_printed += 1
    color = color_names[check.result.status]
    if emojis:
      print(prefix, emoji_chars[check.result.status], end='', sep='')
    print(prefix, colored(check.name, color, attrs=['bold']), sep='', end='')
    if check.result.status != CheckStatus.PASSED:
      print(':', check.result.message)
    else:
      print()

  if num_printed > 0:
    print()

def print_checks_all(
  name: str,
  checks: List[Check],
  seconds: float,
  skip_positive_checks: bool = False,
  print_stats: bool = True,
  use_stderr: bool = False,
):
  package_name = termcolor.colored(name, 'yellow')
  print_checks(checks, prefix='  ', skip_positive_checks=skip_positive_checks, use_stderr=use_stderr)
  if print_stats:
    print('ran', len(checks), 'checks for package', package_name, 'in {:.3f}s'.format(seconds))


def get_checks_status(checks: List[Check], warnings_as_errors: bool = False) -> int:
  max_level = max(x.result.status for x in checks)
  if max_level == CheckStatus.PASSED:
    status = 0
  elif max_level == CheckStatus.WARNING:
    status = 1 if warnings_as_errors else 0
  elif max_level ==  CheckStatus.ERROR:
    status = 1
  else:
    assert False, max_level
  return status



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from shut.model import MonorepoModel
from .base import CheckStatus, CheckResult, Checker, SkipCheck, check, register_checker


class MonorepoChecker(Checker[MonorepoModel]):

  @check('invalid-package')
  def _check_no_invalid_packages(self, monorepo):
    for package_name, exc_info in monorepo.project.invalid_packages:
      yield CheckResult(CheckStatus.ERROR, f'{package_name}: {exc_info[1]}')

  @check('inconsistent-single-version')
  def _check_consistent_mono_version(self, monorepo):
    if monorepo.release.single_version and monorepo.project.packages:
      for package in monorepo.project.packages:
        if package.version is not None and package.version != monorepo.version:
          yield CheckResult(CheckStatus.ERROR, f'{package.name} v{package.version}, expected v{monorepo.version}')
    yield SkipCheck()


register_checker(MonorepoModel, MonorepoChecker)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Get license information from DejaCode.
"""

import json

import click
from shut.utils.external.license import get_license_metadata, wrap_license_text
from . import shut


@shut.group(help=__doc__)
def license():
  pass


@license.command()
@click.option('--name', help='The name of the license to retrieve.', required=True)
@click.option('--long', 'format_', flag_value='long', default=True)
@click.option('--short', 'format_', flag_value='short')
@click.option('--json', 'format_', flag_value='json')
def get(name, format_):
  " Retrieve the license text or a JSON description of the license. "

  data = get_license_metadata(name)
  if format_ == 'json':
    print(json.dumps(data, sort_keys=True))
  elif format_ == 'long':
    print(wrap_license_text(data['license_text']))
  elif format_ == 'short':
    print(wrap_license_text(data['standard_notice'] or data['license_text']))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import enum
import datetime
import json
import os
import shlex
import shutil
import subprocess as sp
from dataclasses import dataclass, field
from typing import Optional, List, TYPE_CHECKING

from shut.model.requirements import Requirement

if TYPE_CHECKING:
  from shut.model.package import PackageModel

__all__ = [
  'TestStatus',
  'TestEnvironment',
  'StackTrace',
  'TestCrashReport',
  'TestCase',
  'TestRun',
  'Runtime',
  'Virtualenv',
  'BaseTestDriver',
]


class TestStatus(enum.Enum):
  PASSED = enum.auto()  #: The test passed.
  FAILED = enum.auto()  #: The test failed.
  SKIPPED = enum.auto()  #: The test was skipped.
  ERROR = enum.auto()  #: An error occurred when running the test.


@dataclass
class TestEnvironment:
  python_version: str
  platform: Optional[str]


@dataclass
class StackTrace:
  filename: str
  lineno: int
  message: str


@dataclass
class TestCrashReport:
  filename: str
  lineno: int
  message: str
  traceback: List[StackTrace]
  longrepr: str


@dataclass
class TestCase:
  name: str
  duration: float
  filename: str
  lineno: int
  status: TestStatus
  crash: Optional[TestCrashReport]
  stdout: str


@dataclass
class TestError:
  filename: Optional[str]
  longrepr: str


@dataclass
class TestRun:
  started: datetime.datetime
  duration: float
  status: TestStatus
  environment: TestEnvironment
  tests: List[TestCase]
  errors: List[TestError] = field(default_factory=list)


@dataclass
class Runtime:
  python: List[str]
  pip: List[str]
  virtualenv: List[str]

  @classmethod
  def from_env(self) -> 'Runtime':
    python = shlex.split(os.getenv('PYTHON', 'python'))
    pip = shlex.split(os.getenv('PIP')) if os.getenv('PIP') else python + ['-m', 'pip']
    virtualenv = shlex.split(os.getenv('VIRTUALENV')) if os.getenv('VIRTUALENV') else python + ['-m', 'venv']
    return Runtime(python, pip, virtualenv)

  @classmethod
  def from_python3(self, python: List[str]) -> 'Runtime':
    pip = python + ['-m', 'pip']
    virtualenv = python + ['-m', 'venv']
    return Runtime(python, pip, virtualenv)

  def get_environment(self) -> TestEnvironment:
    env = getattr(self, '_environment', None)
    if env is None:
      code = 'import sys, platform, json\n'\
            'print(json.dumps({"version": sys.version, "platform": platform.platform()}))'
      raw = json.loads(sp.check_output(self.python + ['-c', code]).decode())
      env =TestEnvironment(raw['version'], raw['platform'])
      self._environment = env
    return env


@dataclass
class Virtualenv:
  path: str

  def exists(self) -> bool:
    return os.path.isdir(self.path)

  def create(self, runtime: Runtime) -> None:
    """
    Create the virtual environment.
    """

    sp.check_call(runtime.virtualenv + [self.path])

  def rm(self):
    if os.path.isdir(self.path):
      shutil.rmtree(self.path)

  def bin(self, name):
    if os.name == 'nt':
      name += '.exe'
    return os.path.join(self.path, 'Scripts' if os.name == 'nt' else 'bin', name)

  def get_runtime(self) -> Runtime:
    return Runtime.from_python3([self.bin('python')])


class BaseTestDriver(metaclass=abc.ABCMeta):
  """
  Base class for drivers that can run unit tests for a package.
  """

  @abc.abstractmethod
  def test_package(self, package: 'PackageModel', runtime: Runtime) -> TestRun:
    pass

  @abc.abstractmethod
  def get_test_requirements(self) -> List[Requirement]:
    pass



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
API for describing sanity checks around a package or monorepo configuration.
"""

from typing import Callable, Iterable, Generic, NamedTuple, Optional, Type, TypeVar
import enum
import types

from shut.model import AbstractProjectModel
from shut.utils.type_registry import TypeRegistry

__all__ = [
  'CheckStatus',
  'CheckResult',
  'Check',
  'register_checker',
  'get_checks',
]

T = TypeVar('T')


class CheckStatus(enum.IntEnum):
  PASSED = enum.auto()  #: The check has passed.
  WARNING = enum.auto()  #: The check is merely giving a warning.
  ERROR = enum.auto()  #: The check has an error, something is in a bad or invalid state.


class CheckResult(NamedTuple):
  """
  Yield this from a #@check() decorated method on a #Checker subclass to return a result
  for the check. Multiple results can be returned from a check.
  """

  status: CheckStatus
  message: str
  subject: Optional[AbstractProjectModel] = None


class SkipCheck(NamedTuple):
  """
  Yield this from a #@check() decorated method on a #Checker subclass to indicate that the
  check should be skipped. This is different from not yielding anything as the check will be
  considered successful in that case.
  """


class Check(NamedTuple):
  name: str
  result: CheckResult


def check(name: str) -> Callable[[Callable], Callable]:
  """
  Decorator for methods on a #Checker instance.
  """

  def decorator(func: Callable) -> Callable:
    func.__check_name__ = name  # type: ignore
    return func

  return decorator


class Checker(Generic[T]):

  def get_checks(self, subject: T) -> Iterable[Check]:
    """
    Yield #Check objects for the *subject*. By default, all methods decorated with
    #check() are called.
    """

    for key in dir(self):
      value = getattr(self, key)
      check_value = value
      if isinstance(value, types.MethodType):
        check_value = value.__func__
      if isinstance(check_value, types.FunctionType) and hasattr(check_value, '__check_name__'):
        index = None
        for index, result in enumerate(value(subject)):
          if not isinstance(result, SkipCheck):
            yield Check(value.__check_name__, result)
        if index is None:
          yield Check(value.__check_name__, CheckResult(CheckStatus.PASSED, None))


registry = TypeRegistry[Checker]()


def register_checker(t: Type[T], checker: Type[Checker[T]]) -> None:
  """
  Register a *checker* class to run checks for type *t*.
  """

  registry.put(t, checker)


def get_checks(obj: T) -> Iterable[Check]:
  """
  Returns all checks from the checkers registered for the type of *obj*.
  """

  for checker in registry.for_type(type(obj)):
    yield from checker().get_checks(obj)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
from typing import Iterable

from shut.model.package import PackageError, PackageModel
from shut.renderers import get_files
from shut.utils.external.classifiers import get_classifiers
from .base import CheckResult, CheckStatus, Checker, SkipCheck, check, register_checker


class PackageChecker(Checker[PackageModel]):

  @check('readme')
  def _check_readme(self, package: PackageModel) -> Iterable[CheckResult]:
    if not package.get_readme_file():
      yield CheckResult(CheckStatus.WARNING, 'No README file found.')

  @check('license')
  def _check_license(self, package: PackageModel) -> Iterable[CheckResult]:
    if not package.get_license_file(True):
      if not package.license:
        yield CheckResult(CheckStatus.WARNING, 'not specified')
      else:
        yield CheckResult(CheckStatus.WARNING, 'No LICENSE file')

    monorepo = package.project.monorepo
    if package.license and monorepo and monorepo.license \
        and monorepo.license != package.license:
      yield CheckResult(CheckStatus.ERROR,
        'License is not consistent with parent mono repository (package: {}, monorepo: {}).'
          .format(package.license, monorepo.license))

  @check('classifiers')
  def _check_classifiers(self, package: PackageModel) -> Iterable[CheckResult]:
    classifiers = get_classifiers()
    unknown_classifiers = [x for x in package.classifiers if x not in classifiers]
    if unknown_classifiers:
      yield CheckResult(
        CheckStatus.WARNING,
        'Unknown classifiers: ' + ', '.join(unknown_classifiers))

  @check('package-url')
  def _check_author(self, package: PackageModel) -> Iterable[CheckResult]:
    if not package.get_url():
      yield CheckResult(CheckStatus.WARNING, 'missing')

  @check('package-author')
  def _check_consistent_author(self, package: PackageModel) -> Iterable[CheckResult]:
    if not package.get_author():
      yield CheckResult(CheckStatus.ERROR, 'missing')
    metadata = package.get_python_package_metadata()
    try:
      author = metadata.author
    except PackageError as exc:
      yield CheckResult(CheckStatus.ERROR, str(exc))
      return
    if package.get_author() and author != str(package.get_author()):
      yield CheckResult(
        CheckStatus.ERROR,
        'Inconsistent package author (package.yaml: {!r} != {}: {!r})'.format(
          str(package.get_author()), metadata.filename, author))

  @check('package-version')
  def _check_consistent_version(self, package: PackageModel) -> Iterable[CheckResult]:
    metadata = package.get_python_package_metadata()
    try:
      version = metadata.version
    except PackageError as exc:
      yield CheckResult(CheckStatus.ERROR, str(exc))
      return
    if package.version and version != str(package.version):
      yield CheckResult(
        CheckStatus.ERROR,
        '{!r} ({}) != {!r} ({})'.format(
          str(package.version),
          os.path.basename(package.filename),
          version,
          os.path.relpath(metadata.filename)))

  @check('typed')
  def _check_typed(self, package: PackageModel) -> Iterable[CheckResult]:
    metadata = package.get_python_package_metadata()
    try:
      py_typed_file = os.path.join(metadata.package_directory, 'py.typed')
    except PackageError as exc:
      yield CheckResult(CheckStatus.ERROR, str(exc))
    except ValueError as exc:
      if package.typed:
        yield CheckResult(
          CheckStatus.WARNING,
          '$.package.typed only works with packages, but this is a module')
    else:
      if os.path.isfile(py_typed_file) and not package.typed:
        yield CheckResult(
          CheckStatus.WARNING,
          'file "py.typed" exists but $.typed is not set')
    yield SkipCheck()

  @check('namespace files')
  def _check_naemspace_files(self, package: PackageModel) -> Iterable[CheckResult]:
    namespaces = package.get_modulename().split('.')[:-1]
    namespaces = [os.sep.join(namespaces[:i+1]) for i in range(len(namespaces))]

    for namespace in namespaces:
      namespace_file = os.path.join(package.get_directory(), package.source_directory, namespace, '__init__.py')
      if not os.path.exists(namespace_file):
        yield CheckResult(CheckStatus.WARNING,
          f'namespace file "{namespace + os.sep}__init__.py" does not exist')

  @check('up to date')
  def _check_up_to_date(self, package: PackageModel) -> Iterable[CheckResult]:
    """
    Checks if the package is up to date.
    """

    files = get_files(package)
    modified_files = files.get_modified_files(package.get_directory())
    if modified_files:
      yield CheckResult(
        CheckStatus.ERROR,
        f'managed files are not up to date ({",".join(modified_files)})')


register_checker(PackageModel, PackageChecker)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from datetime import datetime
from typing import List
import logging
import os
import time

import requests

CACHE_FILENAME = os.path.expanduser('~/.local/shut/classifiers-cache.txt')
CACHE_TTL = 60 * 60 * 24 * 7  # 7 days
CLASSIFIERS_URL = 'https://pypi.org/pypi?%3Aaction=list_classifiers'
logger = logging.getLogger(__name__)
_runtime_cache = None


def get_classifiers(force_refresh: bool = False) -> List[str]:
  """
  Loads the classifiers list from PyPI. Once loaded, the classifiers are cached on disk and
  in memory. The cache on disk is valid for a maxmium of seven days. Specify the *force_refresh*
  argument to ignore any caches.
  """

  global _runtime_cache
  if not force_refresh and _runtime_cache is not None:
    return list(_runtime_cache)

  def _load_cachefile():
    global _runtime_cache
    with open(CACHE_FILENAME) as fp:
      _runtime_cache = [x.rstrip('\n') for x in fp]
    return list(_runtime_cache)

  has_cachefile = not force_refresh and os.path.isfile(CACHE_FILENAME)
  if has_cachefile and (time.time() - os.path.getmtime(CACHE_FILENAME)) < CACHE_TTL:
    return _load_cachefile()
  try:
    classifiers = requests.get(CLASSIFIERS_URL, timeout=1).text.split('\n')
  except requests.exceptions.ReadTimeout as exc:
    if has_cachefile:
      cache_timestamp = datetime.fromtimestamp(os.path.getmtime(CACHE_FILENAME))
      message = ' Falling back to cached classifiers (timestamp: {})'.format(cache_timestamp)
    else:
      message = ' Returning no classifiers. Subsequent classifier validation might raise.'
    logger.warning('Error retrieving classifiers from "%s" (%s).%s', CLASSIFIERS_URL, exc, message)
    if has_cachefile:
      return _load_cachefile()
    _runtime_cache = []
    return []
  try:
    os.makedirs(os.path.dirname(CACHE_FILENAME), exist_ok=True)
    with open(CACHE_FILENAME, 'w') as fp:
      fp.writelines((x + '\n' for x in classifiers))
  except:
    logger.exception('Unable to write classifiers cache file.')

  _runtime_cache = classifiers
  return list(classifiers)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys

import click
from termcolor import colored

from shut.commands.pkg.publish import PublishError, publish_package
from shut.model.monorepo import MonorepoModel
from shut.model.target import TargetId
from . import mono, project


@mono.command()
@click.argument('target', type=lambda s: TargetId.parse(s, True))
@click.option('-t', '--test', is_flag=True, help='publish to the test repository instead')
@click.option('-v', '--verbose', is_flag=True, help='show more output')
@click.option('-b', '--build-dir', default='build', help='build output directory')
@click.option('--skip-build', is_flag=True, help='do not build artifacts that are to be published')
@click.option('-ff', '--fast-fail', is_flag=True, help='Fail after the first error.')
def publish(target, test, verbose, build_dir, skip_build, fast_fail):
  """
  Call `shut pkg publish` for every package. The mono repository must be single versioned
  in order to run this command.
  """

  monorepo = project.load_or_exit(expect=MonorepoModel)
  if not monorepo.release.single_version:
    sys.exit('error: $.release.single-version is not enabled')

  ok = True
  for package in project.packages:
    try:
      publish_package(package, target, build_dir, skip_build, test, verbose)
    except PublishError as exc:
      ok = False
      print(f'error in {colored(package.name, "yellow")}: {exc}', file=sys.stderr)
      if fast_fail:
        sys.exit(1)

  sys.exit(0 if ok else 1)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
The V3 of changelogs.
"""

import enum
from typing import List
from databind.core import datamodel, field
from . import _ChangelogBase, v1


class BumpMode(enum.IntEnum):
  patch = 0
  minor = 1
  major = 2


class Type(enum.Enum):
  fix = ('fix', BumpMode.patch)
  improvement = ('improvement', BumpMode.patch)
  change = ('change', BumpMode.minor)
  breaking_change = ('breaking_change', BumpMode.major)
  refactor = ('refactor', BumpMode.patch)
  feature = ('feature', BumpMode.minor)
  docs = ('docs', BumpMode.patch)
  tests = ('test', BumpMode.patch)

  @property
  def bump_mode(self) -> BumpMode:
    return self.value[1]


@datamodel
class Entry:
  type_: Type = field(altname='type')
  component: str
  description: str
  fixes: List[str]

  Type = Type

  @classmethod
  def from_v1(cls, v1_entry: v1.Entry) -> 'Entry':
    try:
      type_ = Type[v1_entry.types[0].strip().lower()]
    except KeyError:
      type_ = Type.change
    return cls(
      type_=type_,
      component=v1_entry.components[0],
      description=v1_entry.description,
      fixes=list(map(str, v1_entry.issues)),
    )


class Changelog(_ChangelogBase[v1.Changelog], List[Entry]):
  Supersedes = v1.Changelog  # _ChangelogBase

  @classmethod
  def adapt(cls, v1_changelog: v1.Changelog) -> 'Changelog':
    return cls(map(Entry.from_v1, v1_changelog))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import ast
import os
import re
from typing import Dict, List, Optional, Union

from databind.core import datamodel, field
from nr.stream import chain, concat  # type: ignore

from shut.utils.ast import load_module_members
from shut.utils.fs import get_file_in_directory
from .author import Author
from .abstract import AbstractProjectModel
from .linter import LinterConfiguration
from .publish import PublishConfiguration
from .requirements import Requirement, RequirementsList
from .test import TestDriver
from .version import Version


class PackageError(Exception):
  pass


@datamodel
class InstallConfiguration:
  @datamodel
  class InstallHooks:
    before_install: List[str] = field(altname='before-install', default_factory=list)
    after_install: List[str] = field(altname='after-install', default_factory=list)
    before_develop: List[str] = field(altname='before-develop', default_factory=list)
    after_develop: List[str] = field(altname='after-develop', default_factory=list)

    def any(self):
      return any((self.before_install, self.after_install, self.before_develop, self.after_develop))

  hooks: InstallHooks = field(default_factory=InstallHooks)

  #: A value for the `--index-url` option to pass to Pip when using `shut pkg install`.
  index_url: Optional[str] = field(altname='index-url', default=None)

  #: A list of URLs to pass to Pip via the `--extra-index-url` option when using
  #: `shut pkg install`.
  extra_index_urls: List[str] = field(altname='extra-index-urls', default_factory=list)

  def get_pip_args(self) -> List[str]:
    result = []
    if self.index_url:
      result += ['--index-url', self.index_url]
    for url in self.extra_index_urls:
      result += ['--extra-index-url', url]
    return result


@datamodel
class Include:
  include: str


@datamodel
class Exclude:
  exclude: str


@datamodel
class PackageModel(AbstractProjectModel):
  modulename: Optional[str] = None
  description: Optional[str] = None
  readme: Optional[str] = None
  wheel: Optional[bool] = True
  universal: Optional[bool] = None
  typed: Optional[bool] = False
  requirements: RequirementsList = field(default_factory=RequirementsList)
  test_requirements: RequirementsList = field(altname='test-requirements', default_factory=RequirementsList)
  extra_requirements: Dict[str, RequirementsList] = field(altname='extra-requirements', default_factory=dict)
  source_directory: str = field(altname='source-directory', default='src')
  exclude: List[str] = field(default_factory=lambda: ['test', 'tests', 'docs'])
  entrypoints: Dict[str, List[str]] = field(default_factory=dict)
  classifiers: List[str] = field(default_factory=list)
  keywords: List[str] = field(default_factory=list)
  package_data: List[Union[Include, Exclude]] = field(altname='package-data', default_factory=list)

  install: InstallConfiguration = field(default_factory=InstallConfiguration)
  linter: LinterConfiguration = field(default_factory=LinterConfiguration)
  publish: PublishConfiguration = field(default_factory=PublishConfiguration)
  test_driver: TestDriver = field(altname='test-driver', default=None)

  def get_modulename(self) -> str:
    if self.modulename:
      return self.modulename

    # Check for PEP-561 stub packages.
    # See also https://mypy.readthedocs.io/en/latest/installed_packages.html#making-pep-561-compatible-packages
    if self.name.endswith('-stubs'):
      return self.name[:-6].replace('-', '_') + '-stubs'

    return self.name.replace('-', '_')

  def get_python_requirement(self) -> Optional[Requirement]:
    return next(filter(lambda x: isinstance(x, Requirement) and x.package == 'python', self.requirements), None)

  def has_vendored_requirements(self) -> bool:
    """
    Returns #True if the package has any vendored requirements.
    """

    return any(chain(
      self.requirements.vendored_reqs(),
      self.test_requirements.vendored_reqs(),
      concat(l.vendored_reqs() for l in self.extra_requirements.values())
    ))

  def is_universal(self) -> bool:
    """
    Checks if the package is a universal Python package (i.e. it is Python 2 and 3 compatible)
    by testing the `$.requirements.python` version selector. If none is specified, the package
    is also considered universal.
    """

    if self.universal is not None:
      return self.universal

    python_requirement = self.get_python_requirement()
    if not python_requirement:
      return True

    # TODO (@NiklasRosenstein): This method of detecting if the version selector
    #   selects a Python 2 and 3 version is very suboptimal.
    has_2 = re.search(r'\b2\b|\b2\.\b', str(python_requirement))
    has_3 = re.search(r'\b3\b|\b3\.\b', str(python_requirement))
    return bool(has_2 and has_3)

  def get_python_package_metadata(self) -> 'PythonPackageMetadata':
    """
    Returns a #PythonPackageMetadata object for this #PackageModel. This object can be
    used to inspect the author and version information that is defined in the package
    source code.
    """

    return PythonPackageMetadata(
      os.path.join(os.path.dirname(self.filename), self.source_directory),
      self.get_modulename())

  def get_readme_file(self) -> Optional[str]:
    """
    Returns the absolute path to the README for this package.
    """

    directory = os.path.dirname(self.filename)

    if self.readme:
      return os.path.abspath(os.path.join(directory, self.readme))

    return get_file_in_directory(
      directory=directory,
      prefix='README.',
      preferred=['README.md', 'README.rst', 'README.txt', 'README'])

  def get_py_typed_file(self) -> Optional[str]:
    if not self.typed:
      return None

    directory = self.get_python_package_metadata().package_directory
    return os.path.join(directory, 'py.typed')

  def get_publish_config(self) -> PublishConfiguration:
    if self.project and self.project.monorepo and self.project.monorepo.publish:
      return self.project.monorepo.publish
    return self.publish

  def get_license(self) -> Optional[str]:
    if self.license:
      return self.license
    if self.project.monorepo:
      return self.project.monorepo.license
    return None

  def get_author(self) -> Optional[Author]:
    if self.author:
      return self.author
    if self.project.monorepo:
      return self.project.monorepo.author
    return None

  def get_url(self) -> Optional[str]:
    if self.url:
      return self.url
    if self.project.monorepo:
      return self.project.monorepo.url
    return None

  # AbstractProjectModel

  def get_name(self) -> str:
    return self.name

  def get_version(self) -> Optional[Version]:
    if self.version:
      return self.version
    if self.project.monorepo and self.project.monorepo.release.single_version:
      return self.project.monorepo.get_version()
    return None

  def get_tag(self, version: Version) -> str:
    tag_format = self.release.tag_format
    if self.project and self.project.monorepo and '{name}' not in tag_format:
      tag_format = '{name}@' + tag_format
    return tag_format.format(name=self.name, version=version)

  def get_license_file(self, inherit: bool = False) -> Optional[str]:
    if not self.license_file and inherit and self.project.monorepo and \
        (not self.license or self.license == self.project.monorepo.license):
      filename = self.project.monorepo.get_license_file()
      if filename:
        return os.path.join(self.project.monorepo.get_directory(), filename)
    return super().get_license_file(False)


class PythonPackageMetadata:
  """
  Represents the metadata of a Python package on disk.
  """

  def __init__(self, source_directory: str, modulename: str) -> None:
    self.source_directory = source_directory
    self.modulename = modulename
    self._filename = None
    self._author = None
    self._version = None

  @property
  def filename(self) -> str:
    """
    Returns the file that contains the package metadata in the Python source code. This is
    usually the module filename, the package `__init__.py` or `__version__.py`.
    """

    if self._filename:
      return self._filename

    parts = self.modulename.split('.')
    prefix = os.sep.join(parts[:-1])
    choices = [
      parts[-1] + '.py',
      os.path.join(parts[-1], '__version__.py'),
      os.path.join(parts[-1], '__init__.py'),
    ]
    for filename in choices:
      filename = os.path.join(self.source_directory, prefix, filename)
      if os.path.isfile(filename):
        self._filename = filename
        return filename

    raise PackageError('Entry file for package "{}" could not be determined'
                       .format(self.modulename))

  @property
  def package_directory(self) -> str:
    """
    Returns the Python package directory. Raises a #ValueError if this metadata represents
    just a single Python module.
    """

    if self.is_single_module:
      raise ValueError('this package is in module-only form')

    return os.path.dirname(self.filename) or '.'

  @property
  def is_single_module(self) -> bool:
    basename = os.path.basename(self.filename)
    return basename not in ('__init__.py', '__version__.py')

  @property
  def author(self) -> str:
    if not self._author:
      self._load_metadata()
    return self._author

  @property
  def version(self) -> str:
    if not self._version:
      self._load_metadata()
    return self._version

  def _load_metadata(self) -> None:
    members = load_module_members(self.filename)

    author = None
    version = None

    if '__version__' in members:
      try:
        version = ast.literal_eval(members['__version__'])
      except ValueError:
        version = '<Non-literal expression>'

    if '__author__' in members:
      try:
        author = ast.literal_eval(members['__author__'])
      except ValueError:
        author = '<Non-literal expression>'

    self._author = author
    self._version = version



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This package provides the programmatic interface to read, process and write YAML
changelog files.

One changelog file is created per version. The name of the file is the version that
the changelog is for. There may be an "_unreleased.yml" changelog file at any point
in time that contains the logs for the next to-be-released version.

The changelog can be used to determine if the next version of a package requires a
major or minor bump, or if only the patch or build number needs to be incremented.
This is achieved by assigning a keyword to every changelog, and that keyword indicates
the type of change and whether it is breaking an API.
"""

import abc
from typing import Any, Generic, Optional, Type, TypeVar

T = TypeVar('T')


class _ChangelogBase(Generic[T], metaclass=abc.ABCMeta):
  """
  Base class for datamodels subclasses that represent the deserialized form of a changelog in
  a specific version. A newer version should reference the predecessor in the #Supersedes
  class-level attribute and implement the #adapt() method in order to automatically support
  migrating to the next version.
  """

  Supersedes: Optional[Type[T]] = None

  @classmethod
  def migrate(cls, older_changelog: Any) -> '_ChangelogBase':
    if not cls.Supersedes:
      raise TypeError('reached {}, unsure how to migrate from {!r}'.format(
        cls.__module__, type(older_changelog).__name__))
    if not isinstance(older_changelog, cls.Supersedes):
      older_changelog = cls.Supersedes.migrate(older_changelog)
    return cls.adapt(older_changelog)

  @classmethod
  def adapt(cls, older_changelog: T) -> '_ChangelogBase':
    raise NotImplementedError



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import datetime
import os

from pkg_resources import resource_string

from .core import Renderer, register_renderer
from shut.model import AbstractProjectModel, PackageModel
from shut.utils.io.virtual import VirtualFiles


def get_license_template(license_name: str) -> str:
  return resource_string('shut', f'data/license_templates/{license_name}.txt').decode('utf-8')


def has_license_template(license_name: str) -> bool:
  try:
    return get_license_template(license_name)
  except FileNotFoundError:
    pass


class LicenseRenderer(Renderer[AbstractProjectModel]):

  def inherits_monorepo_license(self, package: PackageModel) -> bool:
    """
    Returns #True if the *package* inherits the license file of it's mono repository. This is
    only the case if there is a monorepo, and the package and the monorepo have the same license,
    and the license is one for which we deliver a template for.
    """

    if not package.license:
      if package.project.monorepo and package.project.monorepo.license:
        return True
      return False
    if not has_license_template(package.license):
      return False

    license_file = package.get_license_file()

    # If this is a package and it uses the same license as the monorepo, and the monorepo
    # license exists or is one we would generate, then we skip producing a license for this
    # package.
    if package.project.monorepo and \
        package.project.monorepo.license == package.license and \
        not license_file:
      return True

    return False

  def get_files(self, files: VirtualFiles, model: AbstractProjectModel) -> None:
    if not model.license or (isinstance(model, PackageModel)
        and self.inherits_monorepo_license(model)):
      return

    license_file = model.get_license_file()
    license_file = license_file or os.path.join(model.get_directory(), 'LICENSE.txt')
    files.add_static(
      license_file,
      get_license_template(model.license)
        .format(year=datetime.datetime.utcnow().year, author=model.get_author().name),
    )


register_renderer(AbstractProjectModel, LicenseRenderer)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys
from typing import List

import click
from nr.stream import groupby  # type: ignore
from termcolor import colored

from shut.builders import Builder, get_builders
from shut.model import PackageModel
from shut.model.target import TargetId
from . import pkg
from .. import project


def run_builds(builders: List[Builder], build_dir: str, verbose: bool) -> bool:
  os.makedirs(build_dir, exist_ok=True)
  for builder in builders:
    print(colored(f'building {colored(builder.id, "green")}'))
    for filename in builder.get_outputs():
      print(f'  :: {os.path.join(build_dir, filename)}')
    print()
    success = builder.build(build_dir, verbose)
    if not success:
      print(f'error: building "{builder.id}" failed', file=sys.stderr)
      return False
  return True


@pkg.command()
@click.argument('target', type=lambda s: TargetId.parse(s, True), required=False)
@click.option('-l', '--list', 'list_', is_flag=True, help='list available builders')
@click.option('-b', '--build-dir', default='build', help='build output directory')
@click.option('-v', '--verbose', is_flag=True, help='show more output')
def build(target, list_, build_dir, verbose):
  """
  Produce a build of the package.
  """

  if target and list_:
    sys.exit('error: conflicting options')

  package = project.load_or_exit(expect=PackageModel)
  builders = list(get_builders(package))

  if package.has_vendored_requirements():
    sys.exit(f'error: package has vendored requirements and cannot be built')

  if list_:
    print()
    for scope, builders in groupby(builders, lambda b: b.id.scope):
      print(f'{colored(scope, "green")}:')
      for builder in builders:
        print(f'  {builder.id.name} – {builder.get_description()}')
    print()
    return

  if not target:
    sys.exit('error: no target specified')

  builders = [b for b in builders if target.match(b.id)]
  if not builders:
    sys.exit(f'error: no target matches "{target}"')

  success = run_builds(builders, build_dir, verbose)
  if not success:
    sys.exit(1)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import io
from typing import Iterable, Tuple, Union

SubstRange = Tuple[int, int, str]


def substitute_ranges(text: str, ranges: Iterable[SubstRange], is_sorted: bool = False) -> str:
  """
  Replaces parts of *text* using the specified *ranges* and returns the new text. Ranges
  must not overlap. *is_sorted* can be set to `True` if the input *ranges* are already
  sorted from lowest to highest starting index to optimize the function.
  """

  if not is_sorted:
    ranges = sorted(ranges, key=lambda x: x[0])

  out = io.StringIO()
  max_start_index = 0
  max_end_index = 0
  for index, (istart, iend, subst) in enumerate(ranges):
    if iend < istart:
      raise ValueError(f'invalid range at index {index}: (istart: {istart!r}, iend: {iend!r})')
    if istart < max_end_index:
      raise ValueError(f'invalid range at index {index}: overlap with previous range')

    subst = str(subst)
    out.write(text[max_end_index:istart])
    out.write(subst)
    max_start_index, max_end_index = istart, iend

  out.write(text[max_end_index:])
  return out.getvalue()


def indent_text(text: str, indent: Union[str, int]) -> str:
  """
  Indents the *text* by *indent*.
  """

  if isinstance(indent, int):
    indent = ' ' * indent
  return '\n'.join(indent + l for l in text.splitlines())



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import enum
import os
import posixpath
import re
from typing import Iterable, List, Optional, Union
from urllib.parse import urlparse

from databind.core import datamodel

from .version import bump_version, Version


class VersionSelector(object):
  """
  Represents a version selector that limits which versions of a requirement can be
  picked. It simply wraps a string, but if the string looks like a semver selector,
  it can be converted into a ``setuptools`` compatible version selector string (see
  #to_setuptools()).
  """

  ANY: 'VersionSelector'

  def __init__(self, selector):
    if isinstance(selector, VersionSelector):
      selector = selector._string
    self._string = selector.strip()

  def __str__(self):
    return str(self._string)

  def __repr__(self):
    return 'VersionSelector({!r})'.format(self._string)

  def __eq__(self, other):
    if type(self) == type(other):
      return self._string == other._string
    return False

  def __ne__(self, other):
    return not (self == other)

  def to_setuptools(self):  # type: () -> str
    """
    Converts the version selector to a string that Setuptools/Pip can understand by
    expanding the `~` and `^` range selectors.

    Given a version number X.Y.Z, the selectors will be expanded as follows:

    - `^X.Y.Z` -> `>=X.Y.Z,<X+1.0.0`
    - `~X.Y.Z` -> `>=X.Y.Z,<X.Y+1.0`
    - `X.Y.Z -> ==X.Y.Z`
    """

    # Poor-mans test if this looks like the form 'X.Y.Z' without anything around it.
    if not ',' in self._string and self._string[0].isdigit():
      return '==' + self._string

    regex = r'[~^](\d+\.\d+(\.\d+)?[.\-\w]*)'
    def sub(match):
      index = {'^': 0, '~': 1}[match.group(0)[0]]
      max_version = match.group(1).split('.')[:3]
      if len(max_version) == 2:
        max_version.append('0')
      if '-' in max_version[-1]:
        max_version[-1] = max_version[-1].partition('-')[0]
      max_version[index] = str(int(max_version[index]) + 1)
      for i in range(index+1, 3):
        max_version[i] = '0'
      return '>={},<{}'.format(match.group(1), '.'.join(max_version))

    s = self._string + '.0' * (3 - self._string.count('.') - 1)
    return re.sub(regex, sub, s)

  def is_semver_selector(self) -> bool:
    return self._string and self._string[0] in '^~' and ',' not in self._string

  def matches(self, version: Union[Version, str]) -> bool:
    if not self.is_semver_selector():
      # TODO (@NiklasRosenstein): Match setuptools version selectors.
      return False
    min_version = Version(self._string[1:])
    if self._string[0] == '^':
      max_version = bump_version(min_version, 'major')
    elif self._string[0] == '~':
      max_version = bump_version(min_version, 'minor')
    else:
      raise RuntimeError('invalid semver selector string {!r}'.format(self._string))
    return min_version <= Version(version) < max_version


VersionSelector.ANY = VersionSelector('*')


@datamodel(serialize_as=lambda: Union[Requirement, VendoredRequirement])
class BaseRequirement(metaclass=abc.ABCMeta):

  @classmethod
  @abc.abstractmethod
  def databind_json_load(self, value, context):
    pass

  @abc.abstractmethod
  def databind_json_dump(self, context):
    pass


@datamodel
class Requirement(BaseRequirement):
  """
  A Requirement is simply combination of a package name and a version selector.
  """

  INVALID_CHARACTERS = '/&:;\'"!{}()%'

  package: str
  version: VersionSelector = VersionSelector.ANY
  extras: Optional[List[str]] = None

  def __str__(self, *, setuptools: bool = False):
    result = self.package
    if self.extras:
      result += '[' + ','.join(sorted(self.extras)) + ']'
    if self.version != VersionSelector.ANY:
      result += ' ' + (str(self.version.to_setuptools()) if setuptools else str(self.version))
    return result

  def __repr__(self):
    return repr(str(self))

  @classmethod
  def parse(cls, requirement_string: str) -> 'Requirement':
    error = ValueError('invalid requirement: {!r}'.format(requirement_string))
    if set(requirement_string) & set(cls.INVALID_CHARACTERS):
      raise error
    match = re.match(r'''
      ^\s*                # Allow leading whitespace
      ([\w\d\-\._]+)      # Package name
      (?:\[([^\]]+)\])?   # Package extras
      (?:\s*([^\[\]]+))?  # Version selector (do not allow brackets to avoid matching empty extras)
      $''', requirement_string, re.VERBOSE)
    if not match:
      raise error
    package, extras, version = match.groups()
    return cls(package, VersionSelector(version or VersionSelector.ANY),
      extras=sorted(map(str.strip, extras.split(','))) if extras else None)

  def to_setuptools(self) -> str:
    return self.__str__(setuptools=True)

  @classmethod
  def databind_json_load(cls, value, context):
    if isinstance(value, str):
      try:
        return Requirement.parse(value)
      except ValueError as exc:
        raise context.type_error(str(exc))
    return NotImplemented

  def databind_json_dump(self, context):
    return str(self)


@datamodel
class VendoredRequirement(BaseRequirement):
  """
  A vendored requirement is either a relative path or a string prefixed with `git+` that
  Pip understands as an installable source.
  """

  INVALID_CHARACTERS = '*&^;\'"!{}[]()%'

  class Type(enum.Enum):
    Git = enum.auto()
    Path = enum.auto()

  type: Type
  location: str

  def __post_init__(self):
    if self.type == self.Type.Path:
      self.location = self._normpath(self.location)

  def __str__(self):
    if self.type == self.Type.Git:
      return 'git+' + self.location
    elif self.type == self.Type.Path:
      return self._normpath(self.location)
    assert False, self.type

  @staticmethod
  def _normpath(path: str) -> str:
    """
    Normalizes a path in the way that it should be represented in the serialized form of the
    vendored requirement (as posix path).
    """

    if os.name == 'nt':
      path = path.replace('\\', '/')
    if not posixpath.isabs(path):
      path = './' + posixpath.normpath(path)
    return path

  @classmethod
  def parse(cls, requirement_string: str, fallback_to_path: bool = False) -> 'VendoredRequirement':
    error = ValueError(f'invalid vendored requirement: {requirement_string!r}')
    if set(requirement_string) & set(cls.INVALID_CHARACTERS):
      raise error
    if requirement_string.startswith('git+'):
      url = requirement_string[4:]
      if not urlparse(url).scheme:
        raise error
      return cls(cls.Type.Git, url)
    result = urlparse(requirement_string)
    if result.scheme:
      raise error
    if not requirement_string.startswith('./') and not fallback_to_path:
      raise error
    return cls(cls.Type.Path, requirement_string)

  def to_setuptools(self) -> str:
    raise RuntimeError('VendoredRequirement is not supported in setuptools')

  def get_pip_args(self, root: str, develop: bool) -> List[str]:
    args = [self.location]
    if self.type == self.Type.Path:
      args[0] = os.path.normpath(os.path.join(root, args[0]))
      # We gotta make sure that Pip recognizes the argument as a path if there's no path
      # separator in there.
      if args[0].count(os.sep) == 0:
        args[0] = os.path.join(os.curdir, args[0])
      if develop:
        args.insert(0, '-e')
    return args

  @classmethod
  def databind_json_load(cls, value, context):
    if isinstance(value, str):
      try:
        return VendoredRequirement.parse(value)
      except ValueError as exc:
        raise context.type_error(str(exc))
    raise context.type_error()

  def databind_json_dump(self, context):
    return str(self)


class RequirementsList(List[BaseRequirement]):
  """
  A list of requirements. Provides some convenience methods.
  """

  def reqs(self) -> Iterable[Requirement]:
    return filter(lambda x: isinstance(x, Requirement), self)

  def vendored_reqs(self) -> Iterable[VendoredRequirement]:
    return filter(lambda x: isinstance(x, VendoredRequirement), self)

  def get_pip_args(self, root: str, develop: bool) -> List[str]:
    result = []
    for req in self.reqs():
      result.append(req.to_setuptools())
    for req in self.vendored_reqs():
      result += req.get_pip_args(root, develop)
    return result



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import os
import warnings
from typing import List, Optional

from databind.core import datamodel, field

from .author import Author
from .changelog import ChangelogConfiguration
from .release import ReleaseConfiguration
from .version import Version
from shut.utils.fs import get_file_in_directory


@datamodel
class AbstractProjectModel(metaclass=abc.ABCMeta):
  # Derived fields that are not de-serialized per-se, but may be filled during the
  # deserialization process to track additional metadata.
  project: Optional['Project'] = field(derived=True, default=None)
  filename: Optional[str] = field(derived=True, default=None)
  unknown_keys: List[str] = field(derived=True, default_factory=list)

  name: str
  version: Optional[Version] = None
  author: Optional[Author] = None
  license: Optional[str] = None
  license_file: Optional[str] = None
  url: Optional[str] = None
  changelog: ChangelogConfiguration = field(default_factory=ChangelogConfiguration)
  release: ReleaseConfiguration = field(default_factory=ReleaseConfiguration)

  def get_name(self) -> str:
    warnings.warn(f'{type(self).__name__}.get_name() is deprecated, use .name instead', DeprecationWarning)
    return self.name

  def get_version(self) -> Optional[Version]:
    warnings.warn(f'{type(self).__name__}.get_version() is deprecated, use .version instead', DeprecationWarning)
    return self.version

  def get_tag(self, version: Version) -> str:
    return self.release.tag_format.format(name=self.get_name(), version=version)

  def get_directory(self) -> str:
    assert self.filename
    return os.path.dirname(self.filename)

  def get_author(self) -> Optional[Author]:
    return self.author

  def get_changelog_directory(self) -> str:
    assert self.filename
    return os.path.join(os.path.dirname(self.filename), self.changelog.directory)

  def get_license_file(self, inherit: bool = False) -> Optional[str]:
    """
    Returns the absolute path to the LICENSE file for this package.
    """

    if self.license_file:
      return self.license_file

    return get_file_in_directory(
      directory=os.path.dirname(self.filename),
      prefix='LICENSE.',
      preferred=['LICENSE', 'LICENSE.txt', 'LICENSE.rst', 'LICENSE.md'])


from . import Project  # pylint: disable=unused-import



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import click
import sys

from shut.commands import shut, commons, project
from shut.model import Project, PackageModel


@shut.group()
@click.option('--checks/--no-checks', 'run_checks', default=True,
  help='Run checks before executing the subcommand (default: true)')
@click.pass_context
def pkg(ctx, run_checks):
  """
  Manage the Python package in the current directory.
  """

  if run_checks and ctx.invoked_subcommand not in ('new', 'checks'):
    package = project.load(expect=PackageModel)
    checks.check_package(package, skip_positive_checks=True, print_stats=False, use_stderr=True)


@pkg.command()
def get_version():
  package = project.load(expect=PackageModel)
  if not package.version:
    click.echo('version not defined', err=True)
    sys.exit(1)
  print(package.version)


from . import build
from . import bump
from . import checks
from . import format
from . import install
from . import new
from . import publish
from . import requirements
from . import status
from . import test
from . import update



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import subprocess

from termcolor import colored


def subprocess_trimmed_call(*args, verbose: bool = False, **kwargs) -> int:
  """
  Executes a subprocess and only shows stderr in red and indented by 2 spaces if there
  was any error output. Returns the status code.

  If *verbose* is enabled, the stdout and stderr is not redirected.
  """

  proc = subprocess.Popen(
    *args,
    stdout=None if verbose else subprocess.PIPE,
    stderr=None if verbose else subprocess.PIPE,
    **kwargs)  # type: ignore

  stdout, stderr = proc.communicate()
  if stderr:
    for line in stderr.decode().splitlines():
      if not line:
        continue
      print(f'  {colored(line, "red")}')

  return proc.wait()
